import os
from Bio import SeqIO


# file in
wd                  = '/Users/songweizhi/Desktop/get_fake_bins_wd'
spades_ctg          = '%s/scaffolds_min2000_spades.fa'               % wd
combined_ref        = '%s/combined_refs.fna'                         % wd
fake_bin_folder     = '%s/fake_bins'                                 % wd
iden_cutoff         = 99.5
coverage_q_cutoff   = 90


# tmp files
blast_op         = '%s/spades_scaffolds_min2000_vs_refs.tab'         % wd
blast_op_BestHit = '%s/spades_scaffolds_min2000_vs_refs_BestHit.tab' % wd


# makeblastdb, run blastn and keep best hit
makeblastdb_cmd   = 'makeblastdb -in %s -dbtype nucl -parse_seqids' % combined_ref
blastn_parameters = '-evalue 1e-5 -outfmt "6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore qlen slen" -task blastn -num_threads 4'
blastn_cmd        = 'blastn -query %s -db %s -out %s %s' % (spades_ctg, combined_ref, blast_op, blastn_parameters)
get_best_hit_cmd  = 'BioSAK BestHit -i %s -o %s' % (blast_op, blast_op_BestHit)
# os.system(makeblastdb_cmd)
# os.system(blastn_cmd)
# os.system(get_best_hit_cmd)


ref_id_set = set()
ref_len_dict = {}
total_ref_seq = 0
for seq_record in SeqIO.parse(combined_ref, 'fasta'):
    seq_genome = seq_record.id.split('_')[0]
    ref_id_set.add(seq_genome)
    total_ref_seq += len(seq_record.seq)
    if seq_genome not in ref_len_dict:
        ref_len_dict[seq_genome] = len(seq_record.seq)
    else:
        ref_len_dict[seq_genome] += len(seq_record.seq)


total_matched_ctg = 0
mag_size_dict = {}
ctg_to_genome_dict = {}
for each_hit in open(blast_op_BestHit):
    match_split = each_hit.strip().split('\t')
    query = match_split[0]
    subject = match_split[1]
    subject_genome = subject.split('_')[0]
    iden = float(match_split[2])
    align_len = int(match_split[3])
    query_len = int(match_split[12])
    subject_len = int(match_split[13])
    query_bin_name = '_'.join(query.split('_')[:-1])
    subject_bin_name = '_'.join(subject.split('_')[:-1])
    coverage_q = float(align_len) * 100 / float(query_len)
    coverage_s = float(align_len) * 100 / float(subject_len)
    if (iden >= iden_cutoff) and (coverage_q >= coverage_q_cutoff):
        total_matched_ctg += query_len
        ctg_to_genome_dict[query] = subject_genome
        if subject_genome not in mag_size_dict:
            mag_size_dict[subject_genome] = query_len
        else:
            mag_size_dict[subject_genome] += query_len


print('MAG\tLength(Kbp)\tCompleteness(%)')
for each_ref in sorted([i for i in ref_id_set]):
    ref_len = ref_len_dict[each_ref]
    mag_len = 0
    if each_ref in mag_size_dict:
        mag_len = mag_size_dict[each_ref]
    mag_len_kbp = float("{0:.2f}".format(mag_len/(1024)))
    recovery = mag_len*100/ref_len
    recovery = float("{0:.2f}".format(recovery))
    print('%s\t%s\t%s' % (each_ref, mag_len_kbp, recovery))


if os.path.isdir(fake_bin_folder) is True:
    os.system('rm -r %s' % fake_bin_folder)
os.mkdir(fake_bin_folder)


total_ctg_len = 0
for ctg_record in SeqIO.parse(spades_ctg, 'fasta'):
    total_ctg_len += len(ctg_record.seq)
    if ctg_record.id in ctg_to_genome_dict:
        assignment = ctg_to_genome_dict[ctg_record.id]
        fake_bin_file = '%s/%s.fa' % (fake_bin_folder, assignment)
        fake_bin_file_handle = open(fake_bin_file, 'a')
        fake_bin_file_handle.write('>%s\n' % ctg_record.id)
        fake_bin_file_handle.write('%s\n'  % ctg_record.seq)
        fake_bin_file_handle.close()


total_ref_seq_Mbp       = total_ref_seq/(1024*1024)
total_ref_seq_Mbp       = float("{0:.2f}".format(total_ref_seq_Mbp))
total_ctg_len_Mbp       = total_ctg_len/(1024*1024)
total_ctg_len_Mbp       = float("{0:.2f}".format(total_ctg_len_Mbp))
total_matched_ctg_Mbp   = total_matched_ctg/(1024*1024)
total_matched_ctg_Mbp   = float("{0:.2f}".format(total_matched_ctg_Mbp))
print('Total length of reference genomes: %s Mbp' % total_ref_seq_Mbp)
print('Total length of contigs >= 2000bp: %s Mbp' % total_ctg_len_Mbp)
print('Total length of matched contigs: %s Mbp'   % total_matched_ctg_Mbp)
print('Fake bins exported to: %s'                 % fake_bin_folder)

'''

MAG	Length(Kbp)	Completeness(%)
AS	3884.6	86.46
CA	1937.62	52.9
CG	2525.75	78.15
CP	1603.3	50.41
DA	3579.45	73.44
DG	4286.35	90.4
DM	4132.4	86.83
EC	3325.88	73.37
EV	3633.37	66.34
FA	2114.1	60.08
FP	2052.16	97.0
HB	2.01	0.06
HR	2651.31	84.21
HT	3291.99	87.71
MS	2923.64	80.44
ND	0.0	0.0
NG	3081.81	83.3
NO	3483.43	82.68
OU	289.21	14.43
PS	3302.12	73.5
SB	1512.16	34.72
SP	1697.96	93.86
SR	2348.46	76.16
SS	3697.88	81.36
TC	3189.34	74.98
TR	1310.49	25.67
Total length of reference genomes: 100.13 Mbp
Total length of contigs >= 2000bp: 88.51 Mbp
Total length of matched contigs: 64.31 Mbp
Fake bins exported to: /Users/songweizhi/Desktop/get_fake_bins_wd/fake_bins

'''



# wd = '/Users/songweizhi/Desktop/assess_linkages_hc'
# blastn_bin_vs_ref           = '%s/file_in/cami_hc_bin_vs_ref.tab'                                   % wd
# iden_cutoff                 = 99
# aln_len_cutoff              = 1000
# cov_q_cutoff                = 90
#
# for match in open(blastn_bin_vs_ref):
#     match_split = match.strip().split('\t')
#     query = match_split[0]
#     query_genome = '_'.join(query.split('_')[:3])
#     subject = match_split[1]
#     subject_genome = '_'.join(subject.split('_')[:2])
#     iden = float(match_split[2])
#     aln_len = int(match_split[3])
#     query_len = int(match_split[12])
#     subject_len = int(match_split[13])
#     coverage_q = float(aln_len) * 100 / float(query_len)
#     if (iden >= iden_cutoff) and (aln_len > aln_len_cutoff) and (coverage_q >= cov_q_cutoff):
#         if query_genome == 'cami_hc_3':
#             print(match_split)



wd = '/Users/songweizhi/Desktop/assess_linkages_hc'
matam_16s_blastn            = '%s/file_in/matam_16S_uclust_0.999_vs_ref.tab'                        % wd
iden_cutoff_16s             = 99.8  # 99.3 (best), 99.5
aln_len_cutoff_16s          = 500
cov_q_cutoff_16s            = 70

for match in open(matam_16s_blastn):
    match_split = match.strip().split('\t')
    query = match_split[0]
    subject = match_split[1]
    subject_gnm = subject.split('_16S_')[0]
    iden = float(match_split[2])
    aln_len = int(match_split[3])
    query_len = int(match_split[12])
    subject_len = int(match_split[13])
    coverage_q = float(aln_len) * 100 / float(query_len)
    if (iden >= iden_cutoff_16s) and (aln_len > aln_len_cutoff_16s) and (coverage_q >= cov_q_cutoff_16s):
        if subject_gnm == 'tax_134676':
            print(match_split)







def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def get_cigar_stats(cigar_splitted):

    # aligned_len: M I X =
    # clipping_len: S
    # mismatch_len: X I D
    # mismatch_pct = mismatch_len / aligned_len
    # aligned_pct  = aligned_len  / (aligned_len + clipping_len)
    # clipping_pct = clipping_len / (aligned_len + clipping_len)

    aligned_len = 0
    clipping_len = 0
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get aligned_len
        if each_part_cate in {'M', 'm', 'I', 'i', 'X', 'x', '='}:
            aligned_len += each_part_len

        # get clipping_len
        if each_part_cate in ['S', 's']:
            clipping_len += each_part_len

        # get mismatch_len
        if each_part_cate in {'I', 'i', 'X', 'x', 'D', 'd'}:
            mismatch_len += each_part_len

    aligned_pct  = float("{0:.2f}".format(aligned_len * 100 / (aligned_len + clipping_len)))
    clipping_pct = float("{0:.2f}".format(clipping_len * 100 / (aligned_len + clipping_len)))
    mismatch_pct = float("{0:.2f}".format(mismatch_len * 100 / (aligned_len)))

    return aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct

sam_file_mini_assembly_combined     = '/Users/songweizhi/Desktop/round2/scaffolds_combined.sam'
min_M_pct                           = 20
min_M_len                           = 30
min_clp_len                         = 30
min_clp_M_len                       = 25
max_mis_pct                         = 3
MappingRecord_dict = {}
round_2_MappingRecord_dict = {}


# remove reads mapped to multiple miniassembly? check later
gap_seq_to_reads_dict = {}
for each_read in open(sam_file_mini_assembly_combined):
    if not each_read.startswith('@'):
        each_read_split = each_read.strip().split('\t')
        cigar = each_read_split[5]
        if cigar != '*':
            read_id = each_read_split[0]
            ref_id  = each_read_split[2]
            cigar_splitted = cigar_splitter(cigar)
            aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct = get_cigar_stats(cigar_splitted)
            if (aligned_len >= min_M_len) and (aligned_pct >= min_M_pct) and (mismatch_pct <= (max_mis_pct)):
                if ref_id not in gap_seq_to_reads_dict:
                    gap_seq_to_reads_dict[ref_id] = [read_id]
                else:
                    gap_seq_to_reads_dict[ref_id].append(read_id)

stats_GapFilling_file_16s = ''
stats_GapFilling_file_ctg = ''
stats_GapFilling_file_16s_handle = open(stats_GapFilling_file_16s, 'w')
stats_GapFilling_file_ctg_handle = open(stats_GapFilling_file_ctg, 'w')
stats_GapFilling_file_16s_handle.write('Gap_seq,s16,Number\n')
stats_GapFilling_file_ctg_handle.write('Gap_seq,ctg,Number\n')
for gap_seq in gap_seq_to_reads_dict:
    gap_seq_mapped_reads = gap_seq_to_reads_dict[gap_seq]
    gap_seq_to_16s_linkage_dict = {}
    gap_seq_to_ctg_linkage_dict = {}
    gap_seq_to_gnm_linkage_dict = {}
    for mapped_read in gap_seq_mapped_reads:
        mapped_read_basename = mapped_read[:-2]
        mapped_read_strand   = mapped_read[-1]

        mapped_read_mate_or_clp_refs_16s = set()
        if mapped_read_basename in MappingRecord_dict:
            read_mr_round_1 = MappingRecord_dict[mapped_read_basename]
            if read_mr_round_1.consider_round_2 is True:
                if mapped_read_strand == '1':
                    if read_mr_round_1.consider_r1_unmapped_mate is True:
                        mapped_read_mate_or_clp_refs_16s = mapped_read_mate_or_clp_refs_16s.union(read_mr_round_1.r2_filtered_refs)
                    if read_mr_round_1.consider_r1_clipping_part is True:
                        mapped_read_mate_or_clp_refs_16s = mapped_read_mate_or_clp_refs_16s.union(read_mr_round_1.r1_filtered_refs)
                if mapped_read_strand == '2':
                    if read_mr_round_1.consider_r2_unmapped_mate is True:
                        mapped_read_mate_or_clp_refs_16s = mapped_read_mate_or_clp_refs_16s.union(read_mr_round_1.r1_filtered_refs)
                    if read_mr_round_1.consider_r2_clipping_part is True:
                        mapped_read_mate_or_clp_refs_16s = mapped_read_mate_or_clp_refs_16s.union(read_mr_round_1.r2_filtered_refs)

        for each_mate_or_clp_ref_16s in mapped_read_mate_or_clp_refs_16s:
            if each_mate_or_clp_ref_16s not in gap_seq_to_16s_linkage_dict:
                gap_seq_to_16s_linkage_dict[each_mate_or_clp_ref_16s] = 1
            else:
                gap_seq_to_16s_linkage_dict[each_mate_or_clp_ref_16s] += 1

        mapped_read_mate_or_clp_refs_ctg = set()
        if mapped_read_basename in round_2_MappingRecord_dict:
            read_mr_round_2 = round_2_MappingRecord_dict[mapped_read_basename]
            if mapped_read_strand == '1':
                if read_mr_round_2.consider_r1_unmapped_mate is True:
                    mapped_read_mate_or_clp_refs_ctg = mapped_read_mate_or_clp_refs_ctg.union(read_mr_round_2.r2_filtered_refs)
                if read_mr_round_2.consider_r1_clipping_part is True:
                    mapped_read_mate_or_clp_refs_ctg = mapped_read_mate_or_clp_refs_ctg.union(read_mr_round_2.r1_filtered_refs)
            if mapped_read_strand == '2':
                if read_mr_round_2.consider_r2_unmapped_mate is True:
                    mapped_read_mate_or_clp_refs_ctg = mapped_read_mate_or_clp_refs_ctg.union(read_mr_round_2.r1_filtered_refs)
                if read_mr_round_2.consider_r2_clipping_part is True:
                    mapped_read_mate_or_clp_refs_ctg = mapped_read_mate_or_clp_refs_ctg.union(read_mr_round_2.r2_filtered_refs)

        for each_mate_or_clp_ref_ctg in mapped_read_mate_or_clp_refs_ctg:
            if each_mate_or_clp_ref_ctg not in gap_seq_to_ctg_linkage_dict:
                gap_seq_to_ctg_linkage_dict[each_mate_or_clp_ref_ctg] = 1
            else:
                gap_seq_to_ctg_linkage_dict[each_mate_or_clp_ref_ctg] += 1

        for each_16s in gap_seq_to_16s_linkage_dict:
            stats_GapFilling_file_16s_handle.write('%s,%s,%s\n' % (gap_seq, each_16s, gap_seq_to_16s_linkage_dict[each_16s]))
        for each_ctg in gap_seq_to_ctg_linkage_dict:
            stats_GapFilling_file_ctg_handle.write('%s,%s,%s\n' % (gap_seq, each_ctg, gap_seq_to_ctg_linkage_dict[each_ctg]))

stats_GapFilling_file_16s_handle.close()
stats_GapFilling_file_ctg_handle.close()







import os
import glob
import argparse
from Bio import SeqIO


parser = argparse.ArgumentParser()

parser.add_argument('-g', required=True, help='genome folder')
parser.add_argument('-x', required=True, help='genome extension')
parser.add_argument('-a', required=True, help='abundance file')
parser.add_argument('-d', required=True, type=int, help='sequencing depth')
parser.add_argument('-l', required=True, type=int, help='read length')
parser.add_argument('-i', required=True, type=int, help='insert size')
parser.add_argument('-GemSIM', required=False, action="store_true", help='print GemSIM commands')

args = vars(parser.parse_args())

genome_folder       = args['g']
genome_ext          = args['x']
genome_abundance    = args['a']
sequencing_depth    = args['d']
read_len            = args['l']
insert_size         = args['i']
run_GemSIM          = args['GemSIM']

# genome_folder       = '/Users/songweizhi/Desktop/selected_genomes_renamed_no_plasmid'
# genome_abundance    = '/Users/songweizhi/Desktop/abundance_3.txt'
# genome_ext          = 'fna'
# sequencing_depth    = 25
# read_len            = 150
# insert_size         = 200

'''
python3 /srv/scratch/z5039045/MarkerMAG_wd/z_simulate_reads.py -g /srv/scratch/z5039045/MarkerMAG_wd/genome_selection/selected_genomes_renamed_no_plasmid -x fna -a abundance_even.txt -d 7 -l 300 -i 200



'''


genome_re = '%s/*.%s' % (genome_folder, genome_ext)
genome_list = [os.path.basename(file_name) for file_name in glob.glob(genome_re)]


genome_size_dict = {}
for genome_in in genome_list:
    pwd_genome_file = '%s/%s' % (genome_folder, genome_in)
    current_genome_len = 0
    for seq_record in SeqIO.parse(pwd_genome_file, 'fasta'):
        current_genome_len += len(seq_record.seq)
    genome_size_dict[genome_in] = current_genome_len


genome_abundance_dict = {}
for genome_abund in open(genome_abundance):
    genome_abund_split = genome_abund.strip().split('\t')
    genome_abundance_dict[genome_abund_split[0]] = float(genome_abund_split[1])


genome_size_total = 0
for genome in genome_size_dict:
    genome_size_total += genome_size_dict[genome]


total_reads_len_to_simulate = genome_size_total*sequencing_depth
total_reads_pair_to_simulate = total_reads_len_to_simulate//(2*read_len)


paires_to_sim_dict_by_size_by_abund = {}
paired_reads_to_sim_by_size_by_abund_total = 0
for genome in genome_abundance_dict:
    current_genome_size = genome_size_dict[genome]
    current_genome_abun = genome_abundance_dict[genome]
    paired_reads_to_sim_by_size_by_abund = total_reads_pair_to_simulate * (current_genome_size/genome_size_total) * current_genome_abun
    paires_to_sim_dict_by_size_by_abund[genome] = paired_reads_to_sim_by_size_by_abund
    paired_reads_to_sim_by_size_by_abund_total += paired_reads_to_sim_by_size_by_abund


paires_to_sim_dict_by_size_by_abund_norm = {}
for paires_to_sim in paires_to_sim_dict_by_size_by_abund:
    paires_to_sim_norm = paires_to_sim_dict_by_size_by_abund[paires_to_sim] * (total_reads_pair_to_simulate/paired_reads_to_sim_by_size_by_abund_total)
    paires_to_sim_dict_by_size_by_abund_norm[paires_to_sim] = round(paires_to_sim_norm)


for genome_to_sim in paires_to_sim_dict_by_size_by_abund_norm:

    genome_to_sim_basename = '.'.join(genome_to_sim.split('.')[:-1])
    if run_GemSIM is True:
        Reads_simulator_cmd = 'python /srv/scratch/z5039045/Softwares/GemSIM_v1.6/GemReads.py -r %s -n %s -m ../../phix/phix_2020_i5_p.gzip -o %s_%sx -q 33 -u d -l 151 -p -c\nmv rd.log %s_rd.log' % (genome_to_sim, paires_to_sim_dict_by_size_by_abund_norm[genome_to_sim], genome_to_sim_basename, sequencing_depth, genome_to_sim_basename)

    else:
        Reads_simulator_cmd = 'BioSAK Reads_simulator -r %s/%s -n %s -l %s -i %s -split' % (genome_folder, genome_to_sim, paires_to_sim_dict_by_size_by_abund_norm[genome_to_sim], read_len, insert_size)

    print(Reads_simulator_cmd)


print('total_reads_pair_to_simulate %sx: %s' % (sequencing_depth, total_reads_pair_to_simulate))



import os
import glob
from Bio import SeqIO


ref_16s_seq = '/Users/songweizhi/Desktop/MarkerMAG_wd/2_MBARC26/combined_16S.ffn'
linkage_file_dir = '/Users/songweizhi/Desktop/linkage_files'

linkage_file_re   = '%s/*_identified_linkages_genome_level.txt' % linkage_file_dir
linkage_file_list = [os.path.basename(file_name) for file_name in glob.glob(linkage_file_re)]


gnm_16s_num_dict = {}
for each_seq in SeqIO.parse(ref_16s_seq, 'fasta'):
    seq_id = each_seq.id
    gnm_id = seq_id.split('_')[0]
    if gnm_id not in gnm_16s_num_dict:
        gnm_16s_num_dict[gnm_id] = 1
    else:
        gnm_16s_num_dict[gnm_id] += 1


ref_gnm_list = sorted([i for i in gnm_16s_num_dict])
ref_16s_num_list = []
for each_ref in ref_gnm_list:
    each_ref_linked_16s_num = gnm_16s_num_dict.get(each_ref, 0)
    ref_16s_num_list.append(each_ref_linked_16s_num)
ref_16s_num_list_str = [str(i) for i in ref_16s_num_list]

print('Reference_id\t%s' % '\t'.join(ref_gnm_list))
print('Reference\t%s' % '\t'.join(ref_16s_num_list_str))


for each_linkage_file in sorted(linkage_file_list):

    pwd_linkage_file = '%s/%s' % (linkage_file_dir, each_linkage_file)
    linkage_file_prefix = each_linkage_file.split('_identified_linkages_genome_level')[0]

    linked_gnm_16s_num_dict = {}
    for each_match in open(pwd_linkage_file):
        if not each_match.startswith('MarkerGene,GenomicSeq,Number'):
            match_split = each_match.strip().split('\t')
            id_16s = match_split[0]
            id_16s_gnm = id_16s.split('_')[0]
            id_gnm = match_split[1]
            if id_16s_gnm == id_gnm:
                if id_gnm not in linked_gnm_16s_num_dict:
                    linked_gnm_16s_num_dict[id_gnm] = 1
                else:
                    linked_gnm_16s_num_dict[id_gnm] += 1

    linked_16s_num_list = []
    for each_ref in ref_gnm_list:
        each_ref_linked_16s_num = linked_gnm_16s_num_dict.get(each_ref, 0)
        linked_16s_num_list.append(each_ref_linked_16s_num)
    linked_16s_num_list_str = [str(i) for i in linked_16s_num_list]
    print('%s\t%s' % (linkage_file_prefix, '\t'.join(linked_16s_num_list_str)))



import os
import glob


sample_list = ['Airways', 'Gastrointestinal_tract', 'Oral', 'Skin', 'Urogenital_tract']


for each_sample in sample_list:
    total_abun_file = '/Users/songweizhi/Desktop/ref_total_abun_%s.txt' % each_sample
    abundance_folder = '/Users/songweizhi/Desktop/abundance_%s' % each_sample
    abundance_file_re = '%s/*.tsv' % abundance_folder
    abundance_file_list = glob.glob(abundance_file_re)
    current_sample_ref_set = set()
    ref_total_abun_dict = {}
    for abundance_file in abundance_file_list:
        for each_ref in open(abundance_file):
            each_ref_split = each_ref.strip().split('\t')
            ref_id = each_ref_split[0]
            ref_abun = float(each_ref_split[1])
            if ref_abun > 0:
                current_sample_ref_set.add(ref_id)

                if ref_id not in ref_total_abun_dict:
                    ref_total_abun_dict[ref_id] = ref_abun
                else:
                    ref_total_abun_dict[ref_id] += ref_abun

    total_abun_file_handle = open(total_abun_file, 'w')
    for each_red_abun in ref_total_abun_dict:
        total_abun_file_handle.write('%s\t%s\n' % (each_red_abun, ref_total_abun_dict[each_red_abun]))
    total_abun_file_handle.close()

    print('%s\t%s' % (len(current_sample_ref_set), each_sample))



# wd = '/Users/songweizhi/Desktop/CAMI_oral'
# #wd = '/srv/scratch/z5039045/MarkerMAG_wd/CAMI_Oral/0_raw_files'
# genome_to_id_tsv = '%s/genome_to_id.tsv' % wd
# file_re = '%s/abundance/*.tsv' % wd
# file_list = glob.glob(file_re)
# fake_quality = '%s/oral_799_ref_genomes_fake_quality.txt' % wd
#
#
# genome_id_to_name_dict = {}
# for each_ref in open(genome_to_id_tsv):
#     each_ref_split = each_ref.strip().split('\t')
#     genome_id_to_name_dict[each_ref_split[0]] = each_ref_split[1]
#
#
#
# fake_quality_handle = open(fake_quality, 'w')
# fake_quality_handle.write('genome,completeness,contamination\n')
# ref_id_set = set()
# for each_file in file_list:
#     for each_line in open(each_file):
#         gnm_id = each_line.split('\t')[0]
#         gnm_name = genome_id_to_name_dict.get(gnm_id, 'na')
#         abun = float(each_line.split('\t')[1])
#         if abun > 0:
#             cp_cmd = 'cp 0_raw_files/genomes/%s oral_799_ref_genomes/%s.fa' % (gnm_name, gnm_id)
#             #print(cp_cmd)
#             #os.system(cp_cmd)
#             print(abun)
#             fake_quality_handle.write('%s.fa,%s,%s\n' % (gnm_id, randrange(95, 100), randrange(0, 3)))
#
#             ref_id_set.add(gnm_id)
# print(len(ref_id_set))
#
# fake_quality_handle.close()




import os

wd = '/Users/songweizhi/Desktop/CAMI_high'
wd = '.'
ref_596_genomes   = '%s/goldstandard_high_pool.filtered.profile.stain.596_ref.txt' % wd
source_genome_txt = '%s/source_genomes.txt' % wd


ref_596_genomes_set = set()
for each_ref in open(ref_596_genomes):
    each_ref_split = each_ref.strip().split('\t')
    cami_genome_id = each_ref_split[5]
    ref_596_genomes_set.add(cami_genome_id)

for each_gnm_file in open(source_genome_txt):
    file_name = each_gnm_file.strip()
    gnm_id = ''
    if '_run' in file_name:
        gnm_id = file_name.split('_run')[0]
        #print(gnm_id)
    elif '.gt1kb.' in file_name:
        gnm_id = file_name.split('.gt1kb.')[0]
    else:
        gnm_id = '.'.join(file_name.split('.')[:-1])
    if gnm_id in ref_596_genomes_set:
        rename_cmd = 'cp /srv/scratch/z5039045/MarkerMAG_wd/CAMI_high/source_genomes/%s /srv/scratch/z5039045/MarkerMAG_wd/CAMI_high/ref_gnm_596/%s.fna' % (file_name, gnm_id)
        os.system(rename_cmd)




class MappingRecord:

    def __init__(self):
        self.r1_refs = dict()
        self.r2_refs = dict()


#
# demo_mr_1 = MappingRecord()
#
# demo_mr_1.r1_refs['ref_1'] = '220='
# demo_mr_1.r1_refs['ref_2'] = '111=1X108='
#
#
# print(demo_mr_1.r1_refs)


MappingRecord_dict = {}
demo_sam = '/Users/songweizhi/Desktop/Kelp_SILVA138_id99_assembled_16S_uclust_0.999.sam'
for each_read in open(demo_sam):
    if not each_read.startswith('@'):
        each_read_split = each_read.strip().split('\t')
        cigar = each_read_split[5]
        if cigar != '*':
            read_id = each_read_split[0]
            read_id_base = '.'.join(read_id.split('.')[:-1])
            read_strand = read_id.split('.')[-1]
            ref_id = each_read_split[2]
            ref_pos = each_read_split[3]
            ref_id_with_pos = '%s_pos_%s' % (ref_id, ref_pos)
            read_seq = each_read_split[9]
            #print(each_read.strip())
            #print(read_strand)
            if read_id_base not in MappingRecord_dict:
                MappingRecord_dict[read_id_base] = MappingRecord()

            if read_strand == '1':
                MappingRecord_dict[read_id_base].r1_refs[ref_id_with_pos] = cigar
            if read_strand == '2':
                MappingRecord_dict[read_id_base].r2_refs[ref_id_with_pos] = cigar



#print(MappingRecord_dict)

#print('\n################\n')

for each_mp in MappingRecord_dict:

    print('%s\tr1\t%s' % (each_mp, MappingRecord_dict[each_mp].r1_refs))
    print('%s\tr2\t%s' % (each_mp, MappingRecord_dict[each_mp].r2_refs))
    print()









'''
MappingRecord_example = MappingRecord
MappingRecord_example.r1 = 'aaa'

print(MappingRecord_example.r1)
print(MappingRecord_example.r2)

'''













def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if each_element.isalpha() is True:
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


pwd_samfile = '/Users/songweizhi/Desktop/BH_ER_050417_assembled_16S_uclust_0.999.sam'

all_mapped_reads_set = set()
for each_read in open(pwd_samfile):

    if not each_read.startswith('@'):
        print(each_read.strip())
        each_read_split = each_read.strip().split('\t')
        read_id = each_read_split[0]
        read_id_base = '.'.join(read_id.split('.')[:-1])
        read_strand = read_id.split('.')[-1]
        ref_id = each_read_split[2]
        ref_id_with_prefix = 'MarkerGene__%s' % each_read_split[2]
        ref_pos = int(each_read_split[3])
        cigar = each_read_split[5]
        read_seq = each_read_split[9]
        cigar_splitted = cigar_splitter(cigar)
        read_id_with_ref_pos = '%s__x__%s__x__%s' % (read_id, ref_id, ref_pos)
        all_mapped_reads_set.add(read_id)
        treat_as_full_match = False

        # for perfectly mapped reads, e.g. 150M
        if ('M' in cigar) and (len(cigar_splitted) == 1):
            treat_as_full_match = True

        # for perfectly mapped reads, e.g. 120M-1D-80M
        if len(cigar_splitted) == 3:
            if (cigar_splitted[0][-1] == 'M') and (cigar_splitted[2][-1] == 'M'):
                mismatch_ratio = int(cigar_splitted[1][:-1]) * 100 / len(read_seq)
                M_pct_l = int(cigar_splitted[0][:-1]) * 100 / len(read_seq)
                M_pct_r = int(cigar_splitted[2][:-1]) * 100 / len(read_seq)
                if (M_pct_l >= 20) and (M_pct_r >= 20) and (mismatch_ratio <= 0.5):
                    treat_as_full_match = True

        if ('S' in cigar) and (len(cigar_splitted) == 2):
            cigar_M_len = 0
            cigar_S_len = 0
            if cigar_splitted[0][-1] == 'M':
                cigar_M_len = int(cigar_splitted[0][:-1])
                cigar_S_len = int(cigar_splitted[1][:-1])
            if cigar_splitted[1][-1] == 'M':
                cigar_M_len = int(cigar_splitted[1][:-1])
                cigar_S_len = int(cigar_splitted[0][:-1])

            cigar_M_pct = cigar_M_len * 100 / (cigar_M_len + cigar_S_len)
            cigar_S_pct = cigar_S_len * 100 / (cigar_M_len + cigar_S_len)




import os
import numpy as np
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord


def cigar_splitter(cigar):
    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def get_cigar_stats(cigar_splitted):
    # aligned_len: M I X =
    # clipping_len: S
    # mismatch_len: X I D
    # mismatch_pct = mismatch_len / aligned_len
    # aligned_pct  = aligned_len  / (aligned_len + clipping_len)
    # clipping_pct = clipping_len / (aligned_len + clipping_len)

    aligned_len = 0
    clipping_len = 0
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get aligned_len
        if each_part_cate in {'M', 'm', 'I', 'i', 'X', 'x', '='}:
            aligned_len += each_part_len

        # get clipping_len
        if each_part_cate in ['S', 's']:
            clipping_len += each_part_len

        # get mismatch_len
        if each_part_cate in {'I', 'i', 'X', 'x', 'D', 'd'}:
            mismatch_len += each_part_len

    aligned_pct = float("{0:.2f}".format(aligned_len * 100 / (aligned_len + clipping_len)))
    clipping_pct = float("{0:.2f}".format(clipping_len * 100 / (aligned_len + clipping_len)))
    mismatch_pct = float("{0:.2f}".format(mismatch_len * 100 / (aligned_len)))

    return aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct


def check_both_ends_clipping(cigar_splitted):
    both_ends_clipping = False
    if len(cigar_splitted) >= 3:
        if (cigar_splitted[0][-1] in ['S', 's']) and (cigar_splitted[-1][-1] in ['S', 's']):
            both_ends_clipping = True

    return both_ends_clipping


class MappingRecord:

    def __init__(self):

        #################### overall ####################

        self.qualified_reads = False

        #################### round 1 16s ####################

        self.r1_16s_ref_dict = dict()
        self.r2_16s_ref_dict = dict()

        self.r1_16s_refs_lowest_mismatch = None
        self.r2_16s_refs_lowest_mismatch = None

        self.r1_16s_refs_no_ignored = dict()
        self.r2_16s_refs_no_ignored = dict()
        self.shared_16s_refs_no_ignored = dict()

        self.both_mapped_to_16s = False

        #################### round 1 ctg ####################

        self.r1_ctg_ref_dict = dict()
        self.r2_ctg_ref_dict = dict()

        self.r1_ctg_refs_lowest_mismatch = None
        self.r2_ctg_refs_lowest_mismatch = None

        self.r1_ctg_refs_no_ignored = dict()
        self.r2_ctg_refs_no_ignored = dict()
        self.shared_ctg_refs_no_ignored = dict()

        self.matched_to_ctg = False

        #################### round 2 ####################

        self.qualified_reads_rd2 = False

        self.r1_ctg_ref_dict_rd2 = dict()
        self.r2_ctg_ref_dict_rd2 = dict()

        self.r1_ctg_refs_lowest_mismatch_rd2 = None
        self.r2_ctg_refs_lowest_mismatch_rd2 = None

        #################### round 2 mini_assembly ####################

        self.r1_mini_ref_dict = dict()
        self.r2_mini_ref_dict = dict()

        self.r1_mini_refs_lowest_mismatch = None
        self.r2_mini_refs_lowest_mismatch = None

        self.r1_mini_refs_no_ignored = dict()
        self.r2_mini_refs_no_ignored = dict()
        self.shared_mini_refs_no_ignored = dict()

        #################################################

        self.r1_longest_clp_cigar = ''
        self.r1_longest_clp_falg = ''
        self.r2_longest_clp_cigar = ''
        self.r2_longest_clp_falg = ''

        self.consider_r1_unmapped_mate = False
        self.consider_r1_clipping_part = False
        self.consider_r2_unmapped_mate = False
        self.consider_r2_clipping_part = False

        self.consider_round_2 = False

        self.r1_filtered_refs = set()
        self.r2_filtered_refs = set()
        self.unmapped_r1_refs = set()
        self.unmapped_r2_refs = set()
        self.clipping_r1_refs = set()
        self.clipping_r2_refs = set()
        self.unmapped_r1_refs_with_pos = set()
        self.unmapped_r2_refs_with_pos = set()
        self.clipping_r1_refs_with_pos = set()
        self.clipping_r2_refs_with_pos = set()


def sam_flag_to_rc(flag_value):
    read_rced = 'na'
    if flag_value != '':
        binary_flag = "{0:b}".format(int(flag_value))
        binary_flag_len = len(str(binary_flag))
        binary_flag_polished = '0' * (12 - binary_flag_len) + str(binary_flag)

        if binary_flag_polished[7] == '0':
            read_rced = False
        if binary_flag_polished[7] == '1':
            read_rced = True

    return read_rced


def get_rc(seq_in):
    seq_in_rc = str(SeqRecord(Seq(seq_in)).reverse_complement().seq)
    return seq_in_rc


def get_max_clp_and_index(r1_cigar_list, r2_cigar_list):
    r1_cigar_list_split = [cigar_splitter(i) for i in r1_cigar_list]
    r2_cigar_list_split = [cigar_splitter(i) for i in r2_cigar_list]

    r1_cigar_list_split_only_clp = []
    for each_r1_cigar_split in r1_cigar_list_split:
        clp_len_l = 0
        if each_r1_cigar_split[0][-1] in ['S', 's']:
            clp_len_l = int(each_r1_cigar_split[0][:-1])
        clp_len_r = 0
        if each_r1_cigar_split[-1][-1] in ['S', 's']:
            clp_len_r = int(each_r1_cigar_split[-1][:-1])
        r1_cigar_list_split_only_clp.append([clp_len_l, clp_len_r])

    r2_cigar_list_split_only_clp = []
    for each_r2_cigar_split in r2_cigar_list_split:
        clp_len_l = 0
        if each_r2_cigar_split[0][-1] in ['S', 's']:
            clp_len_l = int(each_r2_cigar_split[0][:-1])
        clp_len_r = 0
        if each_r2_cigar_split[-1][-1] in ['S', 's']:
            clp_len_r = int(each_r2_cigar_split[-1][:-1])
        r2_cigar_list_split_only_clp.append([clp_len_l, clp_len_r])

    cigar_list_split_only_clp_r1_r2 = [r1_cigar_list_split_only_clp, r2_cigar_list_split_only_clp]

    max_value = 0
    max_value_index = ''
    for num_list_1 in cigar_list_split_only_clp_r1_r2[0]:
        if num_list_1[0] > max_value:
            max_value = num_list_1[0]
            max_value_index = 'r1_l'
        if num_list_1[1] > max_value:
            max_value = num_list_1[1]
            max_value_index = 'r1_r'
    for num_list_2 in cigar_list_split_only_clp_r1_r2[1]:
        if num_list_2[0] > max_value:
            max_value = num_list_2[0]
            max_value_index = 'r2_l'
        if num_list_2[1] > max_value:
            max_value = num_list_2[1]
            max_value_index = 'r2_r'

    # get the best cigar
    best_cigar = ''
    if max_value_index == 'r1_l':
        if best_cigar == '':
            for each_cigar in r1_cigar_list:
                if (each_cigar.startswith('%sS' % max_value)) or (each_cigar.startswith('%ss' % max_value)):
                    best_cigar = each_cigar
    elif max_value_index == 'r1_r':
        if best_cigar == '':
            for each_cigar in r1_cigar_list:
                if (each_cigar.endswith('%sS' % max_value)) or (each_cigar.endswith('%ss' % max_value)):
                    best_cigar = each_cigar
    elif max_value_index == 'r2_l':
        if best_cigar == '':
            for each_cigar in r2_cigar_list:
                if (each_cigar.startswith('%sS' % max_value)) or (each_cigar.startswith('%ss' % max_value)):
                    best_cigar = each_cigar
    elif max_value_index == 'r2_r':
        if best_cigar == '':
            for each_cigar in r2_cigar_list:
                if (each_cigar.endswith('%sS' % max_value)) or (each_cigar.endswith('%ss' % max_value)):
                    best_cigar = each_cigar

    return best_cigar, max_value, max_value_index


def check_cigar_all_clp(cigar_list):
    cigar_all_clp = True
    for each_cigar in cigar_list:
        if ('S' not in each_cigar) and ('s' not in each_cigar):
            cigar_all_clp = False

    return cigar_all_clp


def get_min_max_cigar_S(cigar_list):

    # get cigar_S_set
    cigar_S_set = set()
    for each_cigar in cigar_list:
        if ('S' not in each_cigar) and ('s' not in each_cigar):
            cigar_S_set.add(0)
        else:
            cigar_splitted = cigar_splitter(each_cigar)

            if cigar_splitted[0][-1] in ['S', 's']:
                cigar_S_set.add(int(cigar_splitted[0][:-1]))

            if cigar_splitted[-1][-1] in ['S', 's']:
                cigar_S_set.add(int(cigar_splitted[1][:-1]))

    # get min_cigar_S and max_cigar_S
    min_cigar_S = 0
    max_cigar_S = 0
    if len(cigar_S_set) > 0:
        min_cigar_S = min(cigar_S_set)
        max_cigar_S = max(cigar_S_set)

    return min_cigar_S, max_cigar_S


def get_min_mismatch_from_cigar_list(r1_ref_cigar_set, min_M_len):

    mismatch_set_all_cigar = set()
    mismatch_set_long_M_cigars = set()
    for each_cigar in r1_ref_cigar_set:
        aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct = get_cigar_stats(cigar_splitter(each_cigar))
        mismatch_set_all_cigar.add(mismatch_pct)
        if aligned_len >= min_M_len:
            mismatch_set_long_M_cigars.add(mismatch_pct)

    min_mismatch = 'NA'
    if len(mismatch_set_all_cigar) > 0:
        min_mismatch = min(mismatch_set_all_cigar)
        if len(mismatch_set_long_M_cigars) > 0:
            min_mismatch = min(mismatch_set_long_M_cigars)

    return min_mismatch


########################################################################################################################

# do not remove both ends clip alignments, important !!!!!!

wd                                              = '/Users/songweizhi/Desktop/mini_testing'
sam_file_mini_assembly_reformatted              = '%s/scaffolds_bowtie_reformatted.sam'  % wd
rd2_read_to_extract_flanking_16s                = {'MBARC26_20140984'}
to_extract_read_base_rd2_ctg                    = set()

min_M_len_16s                                   = 45
min_M_len_ctg                                   = 45
min_M_len_mini                                  = 45
min_M_pct                                       = 35
mismatch_cutoff                                 = 2
read_to_marker_connector                        = '___r___'
marker_to_ctg_gnm_Key_connector                 = '___M___'
gnm_to_ctg_connector                            = '___C___'
report_interval                                 = 25000
clp_pct_ctg_side_max_num                        = 65
clp_pct_ratio_cutoff                            = 3.5


mini_assembly_to_16s_reads  = '%s/mini_assembly_to_16s_reads.txt' % wd
mini_assembly_to_ctg_reads  = '%s/mini_assembly_to_ctg_reads.txt' % wd
#stats_GapFilling_ctg        = '%s/stats_GapFilling_ctg.txt' % wd


########################################################################################################################

mini_assembly_len_dict = {}
mini_assembly_mp_dict = {}
for each_read in open(sam_file_mini_assembly_reformatted):
    each_read_split = each_read.strip().split('\t')
    if each_read.startswith('@'):
        mini_assembly_id = ''
        mini_assembly_len = 0
        for each_element in each_read_split:
            if each_element.startswith('SN:'):
                mini_assembly_id = each_element[3:]
            if each_element.startswith('LN:'):
                mini_assembly_len = int(each_element[3:])
        mini_assembly_len_dict[mini_assembly_id] = mini_assembly_len
    else:
        cigar = each_read_split[5]
        if cigar != '*':
            read_id = each_read_split[0]
            read_id_base = '.'.join(read_id.split('.')[:-1])
            read_strand = read_id.split('.')[-1]
            ref_id = each_read_split[2]
            ref_pos = int(each_read_split[3])

            if read_id_base not in mini_assembly_mp_dict:
                mini_assembly_mp_dict[read_id_base] = MappingRecord()

            if read_strand == '1':
                if ref_id not in mini_assembly_mp_dict[read_id_base].r1_mini_ref_dict:
                    mini_assembly_mp_dict[read_id_base].r1_mini_ref_dict[ref_id] = {ref_pos: cigar}
                else:
                    mini_assembly_mp_dict[read_id_base].r1_mini_ref_dict[ref_id][ref_pos] = cigar

            if read_strand == '2':
                if ref_id not in mini_assembly_mp_dict[read_id_base].r2_mini_ref_dict:
                    mini_assembly_mp_dict[read_id_base].r2_mini_ref_dict[ref_id] = {ref_pos: cigar}
                else:
                    mini_assembly_mp_dict[read_id_base].r2_mini_ref_dict[ref_id][ref_pos] = cigar

##################################################### parse MappingRecord_dict ####################################################

mini_assembly_to_16s_reads_dict = dict()
mini_assembly_to_ctg_reads_dict = dict()
processed_num = 0
to_extract_read_base = set()
for each_mp in mini_assembly_mp_dict:

    current_mp_record = mini_assembly_mp_dict[each_mp]
    mini_refs_to_ignore = set()

    ########## get lowest mismatch for r1/r2 16s refs ##########

    # get r1_ref_cigar_set
    r1_ref_cigar_set = set()
    for each_pos_dict in current_mp_record.r1_mini_ref_dict.values():
        each_pos_dict_values = {each_pos_dict[i] for i in each_pos_dict}
        r1_ref_cigar_set.update(each_pos_dict_values)

    # get r2_ref_cigar_set
    r2_ref_cigar_set = set()
    for each_pos_dict in current_mp_record.r2_mini_ref_dict.values():
        each_pos_dict_values = {each_pos_dict[i] for i in each_pos_dict}
        r2_ref_cigar_set.update(each_pos_dict_values)

    r1_ref_min_mismatch = get_min_mismatch_from_cigar_list(r1_ref_cigar_set, min_M_len_mini)
    r2_ref_min_mismatch = get_min_mismatch_from_cigar_list(r2_ref_cigar_set, min_M_len_mini)
    mini_assembly_mp_dict[each_mp].r1_mini_refs_lowest_mismatch = r1_ref_min_mismatch
    mini_assembly_mp_dict[each_mp].r2_mini_refs_lowest_mismatch = r2_ref_min_mismatch

    ########## filter r1 mini refs ##########

    r1_mini_refs_passed_qc = {}
    for r1_mini_ref in current_mp_record.r1_mini_ref_dict:
        r1_matched_pos_dict = current_mp_record.r1_mini_ref_dict[r1_mini_ref]

        # one read need to mapped to one 16S only for one time
        if len(r1_matched_pos_dict) > 1:
            mini_refs_to_ignore.add(r1_mini_ref)
        else:
            r1_mini_ref_pos = list(r1_matched_pos_dict.keys())[0]
            r1_mini_ref_cigar = r1_matched_pos_dict[r1_mini_ref_pos]
            r1_mini_ref_cigar_splitted = cigar_splitter(r1_mini_ref_cigar)

            # check both end clip
            both_end_clp = check_both_ends_clipping(r1_mini_ref_cigar_splitted)
            if both_end_clp is True:
                mini_refs_to_ignore.add(r1_mini_ref)
            else:
                # check mismatch
                r1_aligned_len, r1_aligned_pct, r1_clipping_len, r1_clipping_pct, r1_mismatch_pct = get_cigar_stats(r1_mini_ref_cigar_splitted)
                if r1_ref_min_mismatch == 'NA':
                    mini_refs_to_ignore.add(r1_mini_ref)
                elif (r1_mismatch_pct > r1_ref_min_mismatch) or (r1_mismatch_pct > mismatch_cutoff):
                    mini_refs_to_ignore.add(r1_mini_ref)
                else:
                    # check aligned length
                    if r1_aligned_len < min_M_len_mini:
                        mini_refs_to_ignore.add(r1_mini_ref)
                    else:
                        # check if clp in the middle
                        clip_in_middle = False
                        if ('S' in r1_mini_ref_cigar) or ('s' in r1_mini_ref_cigar):
                            clip_in_middle = True
                            if (r1_mini_ref_cigar_splitted[0][-1] in ['S', 's']) and (r1_mini_ref_pos == 1):
                                clip_in_middle = False
                            if (r1_mini_ref_cigar_splitted[-1][-1] in ['S', 's']):
                                if (r1_mini_ref_pos + r1_aligned_len - 1) == mini_assembly_len_dict[r1_mini_ref]:
                                    clip_in_middle = False

                        # exclude the ref if clp in the middle is True
                        if clip_in_middle is True:
                            mini_refs_to_ignore.add(r1_mini_ref)
                        else:
                            r1_mini_refs_passed_qc[r1_mini_ref] = [r1_mini_ref_cigar]

    ########## filter r2 mini refs ##########

    r2_mini_refs_passed_qc = {}
    for r2_mini_ref in current_mp_record.r2_mini_ref_dict:
        r2_matched_pos_dict = current_mp_record.r2_mini_ref_dict[r2_mini_ref]

        # one read need to mapped to one 16S only once
        if len(r2_matched_pos_dict) > 1:
            mini_refs_to_ignore.add(r2_mini_ref)
        else:
            r2_mini_ref_pos = list(r2_matched_pos_dict.keys())[0]
            r2_mini_ref_cigar = r2_matched_pos_dict[r2_mini_ref_pos]
            r2_mini_ref_cigar_splitted = cigar_splitter(r2_mini_ref_cigar)

            # check both end clip
            both_end_clp = check_both_ends_clipping(r2_mini_ref_cigar_splitted)
            if both_end_clp is True:
                mini_refs_to_ignore.add(r2_mini_ref)
            else:
                # check mismatch
                r2_aligned_len, r2_aligned_pct, r2_clipping_len, r2_clipping_pct, r2_mismatch_pct = get_cigar_stats(r2_mini_ref_cigar_splitted)
                if r2_ref_min_mismatch == 'NA':
                    mini_refs_to_ignore.add(r2_mini_ref)
                elif (r2_mismatch_pct > r2_ref_min_mismatch) or (r2_mismatch_pct > mismatch_cutoff):
                    mini_refs_to_ignore.add(r2_mini_ref)
                else:
                    # check aligned length
                    if r2_aligned_len < min_M_len_mini:
                        mini_refs_to_ignore.add(r2_mini_ref)
                    else:
                        # check if clp in the middle
                        clip_in_middle = False
                        if ('S' in r2_mini_ref_cigar) or ('s' in r2_mini_ref_cigar):
                            clip_in_middle = True
                            if (r2_mini_ref_cigar_splitted[0][-1] in ['S', 's']) and (r2_mini_ref_pos == 1):
                                clip_in_middle = False
                            if (r2_mini_ref_cigar_splitted[-1][-1] in ['S', 's']):
                                if (r2_mini_ref_pos + r2_aligned_len - 1) == mini_assembly_len_dict[r2_mini_ref]:
                                    clip_in_middle = False

                        # exclude the ref if clp in the middle is True
                        if clip_in_middle is True:
                            mini_refs_to_ignore.add(r2_mini_ref)
                        else:
                            r2_mini_refs_passed_qc[r2_mini_ref] = [r2_mini_ref_cigar]

    ####################################################################################################

    r1_mini_refs_no_ignored = {key: value for key, value in r1_mini_refs_passed_qc.items() if key not in mini_refs_to_ignore}
    r2_mini_refs_no_ignored = {key: value for key, value in r2_mini_refs_passed_qc.items() if key not in mini_refs_to_ignore}
    shared_mini_ref_dict = {key: [r1_mini_refs_no_ignored[key][0], r2_mini_refs_no_ignored[key][0]] for key in set(r1_mini_refs_no_ignored).intersection(set(r2_mini_refs_no_ignored))}

    # both r1 and r2 mapped to the same mini assembly
    if len(shared_mini_ref_dict) > 0:
        #print('\n---------------------------------\n')
        #print('%s.1\tr1_mini_refs_no_ignored\t%s'   % ((each_mp, r1_mini_refs_no_ignored)))
        #print('%s.2\tr2_mini_refs_no_ignored\t%s'   % ((each_mp, r2_mini_refs_no_ignored)))
        print('%s\tshared_16s_ref_dict\t%s' % ((each_mp, shared_mini_ref_dict)))

        if (each_mp in rd2_read_to_extract_flanking_16s) and (each_mp not in to_extract_read_base_rd2_ctg):
            for each_mini_assembly in shared_mini_ref_dict:
                if each_mini_assembly not in mini_assembly_to_16s_reads_dict:
                    mini_assembly_to_16s_reads_dict[each_mini_assembly] = {each_mp}
                else:
                    mini_assembly_to_16s_reads_dict[each_mini_assembly].add(each_mp)

        elif (each_mp not in rd2_read_to_extract_flanking_16s) and (each_mp in to_extract_read_base_rd2_ctg):
            for each_mini_assembly in shared_mini_ref_dict:
                if each_mini_assembly not in mini_assembly_to_ctg_reads_dict:
                    mini_assembly_to_ctg_reads_dict[each_mini_assembly] = {each_mp}
                else:
                    mini_assembly_to_ctg_reads_dict[each_mini_assembly].add(each_mp)


# write out mini_assembly_to_16s_reads
mini_assembly_to_16s_reads_handle = open(mini_assembly_to_16s_reads, 'w')
for each_mini_assembly in mini_assembly_to_16s_reads_dict:
    mini_assembly_to_16s_reads_handle.write('%s\t%s\n' % (each_mini_assembly, ','.join([i for i in mini_assembly_to_16s_reads_dict[each_mini_assembly]])))
mini_assembly_to_16s_reads_handle.close()

# write out mini_assembly_to_ctg_reads
mini_assembly_to_ctg_reads_handle = open(mini_assembly_to_ctg_reads, 'w')
for each_mini_assembly in mini_assembly_to_ctg_reads_dict:
    mini_assembly_to_ctg_reads_handle.write('%s\t%s\n' % (each_mini_assembly, ','.join([i for i in mini_assembly_to_ctg_reads_dict[each_mini_assembly]])))
mini_assembly_to_ctg_reads_handle.close()






import os


wd = '/Users/songweizhi/Desktop/round_2_by_blast'
round_2_free_living_ctg             = '%s/round2_free_living_ctg.fa'    % wd
round_2_free_living_16s             = '%s/round2_free_living_16s.fa'    % wd
round_2_free_living_blast_result    = '%s/free_living_reads_blastn.tab' % wd
round_2_align_len_cutoff = 50
round_2_iden_cutoff = 99.5
round_2_free_living_16s_to_ctg_connector = '__f__'
round2_free_living_16s_ref_file = '%s/round2_free_living_16s_refs.txt' % wd
round2_free_living_ctg_ref_file = '%s/round2_free_living_ctg_refs.txt' % wd
gnm_ctg_connector = '___'

stats_GapFilling_file = '%s/stats_GapFilling.txt' % wd


blast_parameters = '-evalue 1e-5 -outfmt "6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore qlen slen" -task blastn -num_threads %s' % 4
makeblastdb_cmd  = 'makeblastdb -in %s -dbtype nucl -parse_seqids' % round_2_free_living_ctg
blastn_cmd       = 'blastn -query %s -db %s -out %s %s' % (round_2_free_living_16s, round_2_free_living_ctg, round_2_free_living_blast_result, blast_parameters)
#os.system(makeblastdb_cmd)
#os.system(blastn_cmd)

'''
cat round2_free_living_16s_R1.fa round2_free_living_16s_R2.fa round2_free_living_16s_UP.fa > round2_free_living_16s.fa
cat round2_free_living_ctg_R1.fa round2_free_living_ctg_R2.fa round2_free_living_ctg_UP.fa > round2_free_living_ctg.fa
'''


round2_free_living_16s_ref_dict = {}
for free_living_read_16s in open(round2_free_living_16s_ref_file):
    free_living_read_16s_split = free_living_read_16s.strip().split('\t')
    read_16s_id = free_living_read_16s_split[0]
    read_16s_refs = free_living_read_16s_split[1].split(',')
    round2_free_living_16s_ref_dict[read_16s_id] = read_16s_refs

round2_free_living_ctg_ref_dict = {}
for free_living_read_ctg in open(round2_free_living_ctg_ref_file):
    free_living_read_ctg_split = free_living_read_ctg.strip().split('\t')
    read_ctg_id = free_living_read_ctg_split[0]
    read_ctg_refs = free_living_read_ctg_split[1].split(',')
    round2_free_living_ctg_ref_dict[read_ctg_id] = read_ctg_refs

round_2_min_aln_len = 50
round_2_min_iden = 99.9
round_2_min_cov = 35
m = 0
free_living_16s_to_ctg_linkage_dict = {}
free_living_16s_to_gnm_linkage_dict = {}
for each_hit in open(round_2_free_living_blast_result):
    match_split = each_hit.strip().split('\t')
    query       = match_split[0]
    subject     = match_split[1]
    identity    = float(match_split[2])
    align_len   = int(match_split[3])
    query_len   = int(match_split[12])
    subject_len = int(match_split[13])
    coverage_q  = float(align_len) * 100 / float(query_len)
    coverage_s  = float(align_len) * 100 / float(subject_len)
    qstart      = int(match_split[6])
    qend        = int(match_split[7])
    sstart      = int(match_split[8])
    send        = int(match_split[9])

    if (align_len >= round_2_min_aln_len) and (identity >= round_2_min_iden) and (coverage_q >= round_2_min_cov) and (coverage_s >= round_2_min_cov):

        if ((1 in [qstart, qend]) or (query_len in [qstart, qend])) and ((1 in [sstart, send]) or (subject_len in [sstart, send])):

            print(each_hit.strip())
            print(qstart)
            print(qend)
            print(sstart)
            print(send)
            print()
            m += 1


        query_16s_refs = round2_free_living_16s_ref_dict.get(query, [])
        subject_ctg_refs = round2_free_living_ctg_ref_dict.get(subject, [])
        for each_query_ref in query_16s_refs:
            for each_subject_ref in subject_ctg_refs:

                if each_subject_ref[-2:] in ['_l', '_r']:
                    each_subject_ref = each_subject_ref[:-2]

                subject_ref_gnm = each_subject_ref.split(gnm_ctg_connector)[0]

                q_ref_to_s_ref_key = '%s%s%s' % (each_query_ref, round_2_free_living_16s_to_ctg_connector, each_subject_ref)
                q_ref_to_s_ref_gnm_key = '%s%s%s' % (each_query_ref, round_2_free_living_16s_to_ctg_connector, subject_ref_gnm)

                if q_ref_to_s_ref_key not in free_living_16s_to_ctg_linkage_dict:
                    free_living_16s_to_ctg_linkage_dict[q_ref_to_s_ref_key] = 1
                else:
                    free_living_16s_to_ctg_linkage_dict[q_ref_to_s_ref_key] += 1

                if q_ref_to_s_ref_gnm_key not in free_living_16s_to_gnm_linkage_dict:
                    free_living_16s_to_gnm_linkage_dict[q_ref_to_s_ref_gnm_key] = 1
                else:
                    free_living_16s_to_gnm_linkage_dict[q_ref_to_s_ref_gnm_key] += 1

print(m)



'''

Refined_33	11.06	no	83.3	no	d__Bacteria;p__Myxococcota;c__Polyangia;o__Haliangiales;f__Haliangiaceae;g__;s__

BH_ER_050417_subsample_100_98	d__Bacteria;p__Myxococcota;c__Polyangia;o__Haliangiales;f__Haliangiaceae;g__Haliangium;s__Haliangium_ochraceum
BH_ER_050417_subsample_75_86	d__Bacteria;p__Myxococcota;c__Polyangia;o__Haliangiales;f__Haliangiaceae;g__Haliangium;s__Haliangium_ochraceum

'''

def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if each_element.isalpha() is True:
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


pwd_samfile = '/Users/songweizhi/Desktop/BH_ER_050417_assembled_16S_uclust_0.999.sam'
perfect_match_min_cigar_M_pct = 50
perfect_match_max_cigar_S_pct = 50
mismatch_ratio_max_value = 1


# export clipping mapped reads and perfectly mapped reads
all_mapped_reads_set = set()
clipping_mapped_reads_list = set()
clipping_reads_mapped_part_dict = {}
perfectly_mapped_reads_dict = {}
for each_read in open(pwd_samfile):
    if not each_read.startswith('@'):
        each_read_split = each_read.strip().split('\t')
        cigar = each_read_split[5]
        if cigar != '*':

            read_id = each_read_split[0]
            ref_id = each_read_split[2]

            if ref_id in ['BH_ER_050417_subsample_100_98', 'BH_ER_050417_subsample_75_86']:

                # if read_id in ['Kelp_31369131.1', 'Kelp_31369131.2']:


                read_id_base = '.'.join(read_id.split('.')[:-1])
                read_strand = read_id.split('.')[-1]
                ref_id_with_prefix = 'MarkerGene__%s' % each_read_split[2]
                ref_pos = int(each_read_split[3])
                read_seq = each_read_split[9]
                cigar_splitted = cigar_splitter(cigar)
                read_id_with_ref_pos = '%s__x__%s__x__%s' % (read_id, ref_id, ref_pos)
                all_mapped_reads_set.add(read_id)
                treat_as_full_match = False

                # for perfectly mapped reads, e.g. 150M
                if ('M' in cigar) and (len(cigar_splitted) == 1):
                    treat_as_full_match = True

                # for perfectly mapped reads, e.g. 120M-1D-80M
                if len(cigar_splitted) == 3:
                    if (cigar_splitted[0][-1] == 'M') and (cigar_splitted[2][-1] == 'M'):
                        mismatch_ratio = int(cigar_splitted[1][:-1]) * 100 / len(read_seq)
                        M_pct_l = int(cigar_splitted[0][:-1]) * 100 / len(read_seq)
                        M_pct_r = int(cigar_splitted[2][:-1]) * 100 / len(read_seq)
                        if (M_pct_l >= 20) and (M_pct_r >= 20) and (mismatch_ratio <= 0.5):
                            treat_as_full_match = True

                if ('S' in cigar) and (len(cigar_splitted) == 2):
                    #print(each_read_split)

                    cigar_M_len = 0
                    cigar_S_len = 0
                    if cigar_splitted[0][-1] == 'M':
                        cigar_M_len = int(cigar_splitted[0][:-1])
                        cigar_S_len = int(cigar_splitted[1][:-1])
                    if cigar_splitted[1][-1] == 'M':
                        cigar_M_len = int(cigar_splitted[1][:-1])
                        cigar_S_len = int(cigar_splitted[0][:-1])

                    cigar_M_pct = cigar_M_len * 100 / (cigar_M_len + cigar_S_len)
                    cigar_S_pct = cigar_S_len * 100 / (cigar_M_len + cigar_S_len)
                    # print(cigar_M_pct)
                    # print(cigar_S_pct)

                    # for clipping reads with unmapped part >= min_cigar_S
                    if (cigar_M_pct >= 30) and (cigar_S_pct >= 30):
                        read_seq_left = read_seq[: int(cigar_splitted[0][:-1])]
                        read_seq_right = read_seq[-int(cigar_splitted[1][:-1]):]

                        if cigar_splitted[0][-1] == 'M':

                            # store the match info of mapped part
                            if ('%s_l' % read_id_with_ref_pos) not in clipping_reads_mapped_part_dict:
                                clipping_reads_mapped_part_dict[('%s_l' % read_id_with_ref_pos)] = [ref_id_with_prefix]
                            else:
                                clipping_reads_mapped_part_dict[('%s_l' % read_id_with_ref_pos)].append(ref_id_with_prefix)

                        if cigar_splitted[1][-1] == 'M':

                            # store the match info of mapped part
                            if ('%s_r' % read_id_with_ref_pos) not in clipping_reads_mapped_part_dict:
                                clipping_reads_mapped_part_dict[('%s_r' % read_id_with_ref_pos)] = [ref_id_with_prefix]
                            else:
                                clipping_reads_mapped_part_dict[('%s_r' % read_id_with_ref_pos)].append(ref_id_with_prefix)

                        clipping_mapped_reads_list.add(read_id)

                    # for clipping reads with unmapped part < min_cigar_S
                    # treat these reads as perfectly mapped reads
                    if (cigar_M_pct >= perfect_match_min_cigar_M_pct) and (cigar_S_pct < perfect_match_max_cigar_S_pct):
                        if read_id_base not in perfectly_mapped_reads_dict:
                            perfectly_mapped_reads_dict[read_id_base] = {read_strand: [ref_id_with_prefix]}
                        else:
                            if read_strand not in perfectly_mapped_reads_dict[read_id_base]:
                                perfectly_mapped_reads_dict[read_id_base][read_strand] = [ref_id_with_prefix]
                            else:
                                perfectly_mapped_reads_dict[read_id_base][read_strand].append(ref_id_with_prefix)

                elif ('S' in cigar) and (len(cigar_splitted) > 2):

                    if ((cigar_splitted[0][-1] == 'M') and (cigar_splitted[-1][-1] == 'S')) or (
                            (cigar_splitted[0][-1] == 'S') and (cigar_splitted[-1][-1] == 'M')):

                        if len(cigar_splitted) == 4:

                            # e.g. 30M-1D-50M-40S
                            if (cigar_splitted[0][-1] == 'M') and (cigar_splitted[2][-1] == 'M') and (
                                    cigar_splitted[3][-1] == 'S'):
                                mismatch_ratio = int(cigar_splitted[1][:-1]) * 100 / len(read_seq)
                                if (mismatch_ratio <= mismatch_ratio_max_value) and (int(cigar_splitted[2][:-1]) >= 30):
                                    cigar_M_pct = (int(cigar_splitted[0][:-1]) + int(cigar_splitted[2][:-1])) * 100 / len(
                                        read_seq)
                                    cigar_S_pct = (int(cigar_splitted[3][:-1])) * 100 / len(read_seq)
                                    if (cigar_M_pct >= perfect_match_min_cigar_M_pct) and (
                                            cigar_S_pct < perfect_match_max_cigar_S_pct):
                                        treat_as_full_match = True

                            # e.g. 40S-30M-1D-50M
                            if (cigar_splitted[0][-1] == 'S') and (cigar_splitted[1][-1] == 'M') and (
                                    cigar_splitted[3][-1] == 'M'):
                                mismatch_ratio = int(cigar_splitted[2][:-1]) * 100 / len(read_seq)
                                if (mismatch_ratio <= mismatch_ratio_max_value) and (int(cigar_splitted[1][:-1]) >= 30):
                                    cigar_M_pct = (int(cigar_splitted[1][:-1]) + int(cigar_splitted[3][:-1])) * 100 / len(
                                        read_seq)
                                    cigar_S_pct = (int(cigar_splitted[0][:-1])) * 100 / len(read_seq)
                                    if (cigar_M_pct >= perfect_match_min_cigar_M_pct) and (
                                            cigar_S_pct < perfect_match_max_cigar_S_pct):
                                        treat_as_full_match = True

                        # elif len(cigar_splitted) == 6:
                        #     if (cigar_splitted[0][-1] == 'M') and (cigar_splitted[2][-1] == 'M') and (cigar_splitted[4][-1] == 'M') and (cigar_splitted[5][-1] == 'S'):
                        #         mismatch_ratio = (int(cigar_splitted[1][:-1]) + int(cigar_splitted[3][:-1])) * 100 / len(read_seq)
                        #         if mismatch_ratio <= mismatch_ratio_max_value:
                        #             cigar_M_pct = (int(cigar_splitted[0][:-1]) + int(cigar_splitted[2][:-1]) + int(cigar_splitted[4][:-1])) * 100 / len(read_seq)
                        #             cigar_S_pct = (int(cigar_splitted[5][:-1])) * 100 / len(read_seq)
                        #             if (cigar_M_pct >= perfect_match_min_cigar_M_pct) and (
                        #                     cigar_S_pct < perfect_match_max_cigar_S_pct):
                        #                 treat_as_full_match = True
                        #
                        #     if (cigar_splitted[0][-1] == 'S') and (cigar_splitted[1][-1] == 'M') and (cigar_splitted[3][-1] == 'M') and (cigar_splitted[5][-1] == 'M'):
                        #         mismatch_ratio = (int(cigar_splitted[2][:-1]) + int(cigar_splitted[4][:-1])) * 100 / len(read_seq)
                        #         if mismatch_ratio <= mismatch_ratio_max_value:
                        #             cigar_M_pct = (int(cigar_splitted[1][:-1]) + int(cigar_splitted[3][:-1]) + int(cigar_splitted[5][:-1])) * 100 / len(read_seq)
                        #             cigar_S_pct = (int(cigar_splitted[0][:-1])) * 100 / len(read_seq)
                        #             if (cigar_M_pct >= perfect_match_min_cigar_M_pct) and (cigar_S_pct < perfect_match_max_cigar_S_pct):
                        #                 treat_as_full_match = True

                if treat_as_full_match is True:
                    if read_id_base not in perfectly_mapped_reads_dict:
                        perfectly_mapped_reads_dict[read_id_base] = {read_strand: [ref_id_with_prefix]}
                    else:
                        if read_strand not in perfectly_mapped_reads_dict[read_id_base]:
                            perfectly_mapped_reads_dict[read_id_base][read_strand] = [ref_id_with_prefix]
                        else:
                            perfectly_mapped_reads_dict[read_id_base][read_strand].append(ref_id_with_prefix)


########################################## extract reads with multiprocessing ##########################################

perfectly_mapped_read_singleton_dict = {}
for perfectly_mapped_read in perfectly_mapped_reads_dict:
    current_value = perfectly_mapped_reads_dict[perfectly_mapped_read]
    if len(current_value) == 1:
        perfectly_mapped_read_singleton_dict[perfectly_mapped_read] = current_value

for each in perfectly_mapped_read_singleton_dict:
    print('%s\t%s' % (each, perfectly_mapped_read_singleton_dict[each]))

print(len(perfectly_mapped_read_singleton_dict))

# Kelp_31369131.2
print(perfectly_mapped_reads_dict.get('Kelp_31369131'))

'''

Kelp_43912804.2
'''


import os

# extract path to the config file
pwd_config_file = os.path.realpath(__file__)
config_file_path = '/'.join(pwd_config_file.split('/')[:-1])

# specify full path to corresponding executables at the right side of colon
config_dict = {'config_file_path'       : config_file_path,
               'bowtie2'                : 'bowtie2',
               'bowtie2_build'          : 'bowtie2-build',
               'samtools'               : 'samtools',
               'blastn'                 : 'blastn',
               'makeblastdb'            : 'makeblastdb',
               'spades'                 : 'spades.py',
               'get_sankey_plot_R'      : '%s/get_sankey_plot.R' % config_file_path
               }




markermag_linkage = '/Users/songweizhi/Desktop/CAMI2_HMP_combined_linkages.txt'
GI_bin_depth_file = '/Users/songweizhi/Desktop/GI_bin_mean_depth.txt'

linked_mag_set = set()
for each_linkage in open(markermag_linkage):
    each_linkage_split = each_linkage.strip().split('\t')
    linked_mag = each_linkage_split[1]
    linked_mag_set.add(linked_mag)
print(linked_mag_set)

for mag_depth in open(GI_bin_depth_file):
    mag_depth_split = mag_depth.strip().split('\t')
    mag_id = '.'.join(mag_depth_split[0].split('.')[:-1])

    if mag_id in linked_mag_set:
        print('%s\tyes' % (mag_depth.strip()))
    else:
        print('%s\tno' % (mag_depth.strip()))

print(len(linked_mag_set))


import os
from Bio import SeqIO
from datetime import datetime


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def report_and_log(message_for_report, log_file, keep_quiet):

    time_format = '[%Y-%m-%d %H:%M:%S]'
    with open(log_file, 'a') as log_handle:
        log_handle.write('%s %s\n' % ((datetime.now().strftime(time_format)), message_for_report))

    if keep_quiet is False:
        print('%s %s' % ((datetime.now().strftime(time_format)), message_for_report))


def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def get_cigar_matched_and_mismatch_pct(cigar):

    cigar_splitted = cigar_splitter(cigar)

    total_len_with_s = 0
    total_len_without_s = 0
    matched_seq_len = 0
    mismatched_seq_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get total len with/without s
        if each_part_cate in {'M', 'I', '=', 'X'}:
            total_len_with_s += each_part_len
            total_len_without_s += each_part_len

        # get total len with s
        if each_part_cate == 'S':
            total_len_with_s += each_part_len

        # get matched part len
        if each_part_cate == '=':
            matched_seq_len += each_part_len

        # get mismatched part len
        if each_part_cate in {'I', 'X', 'D'}:
            mismatched_seq_len += each_part_len

    matched_pct    = float("{0:.2f}".format(matched_seq_len*100/total_len_with_s))
    mismatch_pct = float("{0:.2f}".format(mismatched_seq_len*100/total_len_without_s))

    return matched_pct, mismatch_pct


def get_free_living_mate(ref_in, reads_r1, reads_r2, end_seq_len, max_gap_to_end, num_threads, pwd_bbmap_exe, bbmap_memory, perfect_match_min_cigar_M_pct, global_max_mismatch_pct, pwd_log_file, keep_quiet):

    ref_in_path, ref_in_basename, ref_in_ext = sep_path_basename_ext(ref_in)

    ref_subset      = '%s/%s_ends_%sbp%s'                % (ref_in_path, ref_in_basename, end_seq_len, ref_in_ext)
    sam_file        = '%s/%s_ends_%sbp.sam'              % (ref_in_path, ref_in_basename, end_seq_len)
    bbmap_stderr    = '%s/%s_ends_%sbp_bbmap_stderr.txt' % (ref_in_path, ref_in_basename, end_seq_len)

    # get ref seqs subset
    report_and_log(('Round 2: map reads to %s, get ends sequences' % ref_in_basename), pwd_log_file, keep_quiet)
    ref_subset_handle = open(ref_subset, 'w')
    for ref_seq in SeqIO.parse(ref_in, 'fasta'):

        ref_seq_id = ref_seq.id
        ref_seq_len = len(ref_seq.seq)

        if ref_seq_len < end_seq_len * 2:
            ref_subset_handle.write('>%s\n' % ref_seq_id)
            ref_subset_handle.write('%s\n' % ref_seq.seq)
        else:
            ref_seq_left_end_id = '%s_l' % ref_seq_id
            ref_seq_right_end_id = '%s_r' % ref_seq_id
            ref_seq_left_end = ref_seq.seq[:end_seq_len]
            ref_seq_right_end = ref_seq.seq[-end_seq_len:]

            # write out left end
            ref_subset_handle.write('>%s\n' % ref_seq_left_end_id)
            ref_subset_handle.write('%s\n' % ref_seq_left_end)

            # write out right end
            ref_subset_handle.write('>%s\n' % ref_seq_right_end_id)
            ref_subset_handle.write('%s\n' % ref_seq_right_end)
    ref_subset_handle.close()

    # mapping with bbmap
    report_and_log(('Round 2: map reads to %s, mapping' % ref_in_basename), pwd_log_file, keep_quiet)
    bbmap_parameter_round2 = 'local=t nodisk=t ambiguous=all keepnames=t saa=f trd=t silent=true threads=%s -Xmx%sg' % (num_threads, bbmap_memory)
    bbmap_cmd_round2 = '%s ref=%s in=%s in2=%s outm=%s %s 2> %s' % (pwd_bbmap_exe, ref_subset, reads_r1, reads_r2, sam_file, bbmap_parameter_round2, bbmap_stderr)
    os.system(bbmap_cmd_round2)

    # parse sam file
    report_and_log(('Round 2: map reads to %s, filter sam' % ref_in_basename), pwd_log_file, keep_quiet)
    all_mapped_reads = set()
    qualified_reads_dict = {}
    qualified_reads_to_ref_dict = {}
    ref_in_sam_len_dict = {}
    for each_line in open(sam_file):
        each_line_split = each_line.strip().split('\t')
        if each_line.startswith('@'):
            ref_id = ''
            ref_len = 0
            for each_element in each_line_split:
                if each_element.startswith('SN:'):
                    ref_id = each_element[3:]
                if each_element.startswith('LN:'):
                    ref_len = int(each_element[3:])
            ref_in_sam_len_dict[ref_id] = ref_len
        else:
            cigar = each_line_split[5]
            if cigar != '*':
                read_id = each_line_split[0]
                read_id_base = '.'.join(read_id.split('.')[:-1])
                read_strand = read_id.split('.')[-1]
                ref_id = each_line_split[2]
                ref_len = ref_in_sam_len_dict[ref_id]
                ref_pos = int(each_line_split[3])
                cigar_splitted = cigar_splitter(cigar)
                cigar_match_pct, cigar_mismatch_pct = get_cigar_matched_and_mismatch_pct(cigar)
                all_mapped_reads.add(read_id)

                # get aligned length
                aligned_len = 0
                for each_section in cigar_splitted:
                    each_section_len = int(each_section[:-1])
                    each_section_cate = each_section[-1]
                    if each_section_cate in {'D', '=', 'X'}:
                        aligned_len += each_section_len

                # filter mapped reads
                qualified_mapping = False
                if (cigar_match_pct >= perfect_match_min_cigar_M_pct) and (cigar_mismatch_pct <= global_max_mismatch_pct):

                    # check left end for contig's left end sequence
                    if ref_id[-2:] == '_l':
                        if ref_pos <= max_gap_to_end:
                            qualified_mapping = True

                    # check right end for contig's right end sequence
                    elif ref_id[-2:] == '_r':
                        if cigar_splitted[0][-1] != 'S':
                            ref_pos_end = ref_pos + aligned_len
                            if (ref_len - ref_pos_end) <= max_gap_to_end:
                                qualified_mapping = True

                    # check both ends for contigs without subset
                    else:
                        if (ref_pos <= max_gap_to_end) or ((ref_len - ref_pos - aligned_len) <= max_gap_to_end):
                            qualified_mapping = True

                # proceed with qualified mappings
                if qualified_mapping is True:
                    qualified_reads_to_ref_dict[read_id] = ref_id

                    if read_id_base not in qualified_reads_dict:
                        qualified_reads_dict[read_id_base] = [read_strand]
                    else:
                        qualified_reads_dict[read_id_base].append(read_strand)

    reads_to_extract_to_ref_dict = {}
    for qualified_read in qualified_reads_dict:
        read_strand = qualified_reads_dict[qualified_read]
        if len(read_strand) == 1:
            mapped_mate = ''
            mate_to_extract = ''
            if read_strand == ['1']:
                mapped_mate = '%s.1' % (qualified_read)
                mate_to_extract = '%s.2' % (qualified_read)
            if read_strand == ['2']:
                mapped_mate = '%s.2' % (qualified_read)
                mate_to_extract = '%s.1' % (qualified_read)

            if mate_to_extract not in all_mapped_reads:
                reads_to_extract_to_ref_dict[mate_to_extract] = qualified_reads_to_ref_dict[mapped_mate]

    return reads_to_extract_to_ref_dict




def cigar_splitter(cigar):
    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def get_cigar_matched_and_mismatch_pct(cigar_splitted):

    # aligned_len: M I X =
    # clipping_len: S
    # mismatch_len: X I D
    # mismatch_pct = mismatch_len / aligned_len
    # aligned_pct  = aligned_len  / (aligned_len + clipping_len)
    # clipping_pct = clipping_len / (aligned_len + clipping_len)

    aligned_len = 0
    clipping_len = 0
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get aligned_len
        if each_part_cate in {'M', 'm', 'I', 'i', 'X', 'x', '='}:
            aligned_len += each_part_len

        # get clipping_len
        if each_part_cate in ['S', 's']:
            clipping_len += each_part_len

        # get mismatch_len
        if each_part_cate in {'I', 'i', 'X', 'x', 'D', 'd'}:
            mismatch_len += each_part_len

    aligned_pct  = float("{0:.2f}".format(aligned_len * 100 / (aligned_len + clipping_len)))
    clipping_pct = float("{0:.2f}".format(clipping_len * 100 / (aligned_len + clipping_len)))
    mismatch_pct = float("{0:.2f}".format(mismatch_len * 100 / (aligned_len)))

    return aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct


cigar_in = '45=10X45=100S'
aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct = get_cigar_matched_and_mismatch_pct(cigar_splitter(cigar_in))

print(aligned_len)
print(aligned_pct)
print(clipping_len)
print(clipping_pct)
print(mismatch_pct)




from Bio import SeqIO


def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def check_both_ends_clipping(cigar_splitted):

    both_ends_clipping = False
    if len(cigar_splitted) >= 3:
        if (cigar_splitted[0][-1] in ['S', 's']) and (cigar_splitted[-1][-1] in ['S', 's']):
            both_ends_clipping = True

    return both_ends_clipping


def r12_16s_ref_dict_to_str(r2_16s_ref_dict):

    top_dict_str = 'na'

    if r2_16s_ref_dict != {}:
        top_dict_list = []
        for sub_dict_key in r2_16s_ref_dict:
            sub_dict_value = r2_16s_ref_dict[sub_dict_key]
            bottom_dict_list = []
            for bottom_dict_key in sub_dict_value:
                bottom_dict_value = sub_dict_value[bottom_dict_key]
                bottom_dict_list.append('%s:%s' % (bottom_dict_key, bottom_dict_value))
            bottom_dict_str = '%s' % (';'.join(bottom_dict_list))
            top_dict_list.append('%s:::%s' % (sub_dict_key, bottom_dict_str))

        top_dict_str = ';;;'.join(top_dict_list)

    return top_dict_str


def no_ignored_dict_to_str(r2_16s_refs_no_ignored_dict):
    top_dict_str = 'na'

    if r2_16s_refs_no_ignored_dict != {}:
        top_dict_list = []
        for each_ref in r2_16s_refs_no_ignored_dict:
            each_ref_cigar_list = r2_16s_refs_no_ignored_dict[each_ref]
            each_ref_cigar_str = ','.join(each_ref_cigar_list)
            top_dict_list.append('%s:%s' % (each_ref, each_ref_cigar_str))
        top_dict_str = ';'.join(top_dict_list)

    return top_dict_str


def get_min_mismatch_from_cigar_list(r1_ref_cigar_set, min_M_len):

    mismatch_set_all_cigar = set()
    mismatch_set_long_M_cigars = set()
    for each_cigar in r1_ref_cigar_set:
        aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct = get_cigar_stats(cigar_splitter(each_cigar))
        mismatch_set_all_cigar.add(mismatch_pct)
        if aligned_len >= min_M_len:
            mismatch_set_long_M_cigars.add(mismatch_pct)

    min_mismatch = 'NA'
    if len(mismatch_set_all_cigar) > 0:
        min_mismatch = min(mismatch_set_all_cigar)
        if len(mismatch_set_long_M_cigars) > 0:
            min_mismatch = min(mismatch_set_long_M_cigars)

    return min_mismatch


def get_cigar_stats(cigar_splitted):

    # aligned_len: M I X =
    # clipping_len: S
    # mismatch_len: X I D
    # mismatch_pct = mismatch_len / aligned_len
    # aligned_pct  = aligned_len  / (aligned_len + clipping_len)
    # clipping_pct = clipping_len / (aligned_len + clipping_len)

    aligned_len = 0
    clipping_len = 0
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get aligned_len
        if each_part_cate in {'M', 'm', 'I', 'i', 'X', 'x', '='}:
            aligned_len += each_part_len

        # get clipping_len
        if each_part_cate in ['S', 's']:
            clipping_len += each_part_len

        # get mismatch_len
        if each_part_cate in {'I', 'i', 'X', 'x', 'D', 'd'}:
            mismatch_len += each_part_len

    aligned_pct  = float("{0:.2f}".format(aligned_len * 100 / (aligned_len + clipping_len)))
    clipping_pct = float("{0:.2f}".format(clipping_len * 100 / (aligned_len + clipping_len)))
    mismatch_pct = float("{0:.2f}".format(mismatch_len * 100 / (aligned_len)))

    return aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct


def parse_sam16s_worker(argument_list):

    sorted_sam          = argument_list[0]
    MappingRecord_file  = argument_list[1]
    min_M_len_16s       = argument_list[2]
    mismatch_cutoff     = argument_list[3]
    marker_len_dict     = argument_list[4]

    MappingRecord_file_handle = open(MappingRecord_file, 'w')
    current_read_base = ''
    current_read_base_r1_16s_ref_dict = dict()
    current_read_base_r2_16s_ref_dict = dict()
    with open(sorted_sam) as sorted_sam_opened:
        for each_read in sorted_sam_opened:
            if not each_read.startswith('@'):
                each_read_split = each_read.strip().split('\t')
                cigar = each_read_split[5]
                read_id = each_read_split[0]
                read_id_base = '.'.join(read_id.split('.')[:-1])
                read_strand = read_id.split('.')[-1]
                ref_id = each_read_split[2]
                ref_pos = int(each_read_split[3])

                if current_read_base == '':
                    current_read_base = read_id_base

                    if cigar != '*':
                        if read_strand == '1':
                            current_read_base_r1_16s_ref_dict[ref_id] = {ref_pos: cigar}
                        if read_strand == '2':
                            current_read_base_r2_16s_ref_dict[ref_id] = {ref_pos: cigar}

                elif read_id_base == current_read_base:

                    if cigar != '*':
                        if read_strand == '1':
                            if ref_id not in current_read_base_r1_16s_ref_dict:
                                current_read_base_r1_16s_ref_dict[ref_id] = {ref_pos: cigar}
                            else:
                                current_read_base_r1_16s_ref_dict[ref_id][ref_pos] = cigar
                        if read_strand == '2':
                            if ref_id not in current_read_base_r2_16s_ref_dict:
                                current_read_base_r2_16s_ref_dict[ref_id] = {ref_pos: cigar}
                            else:
                                current_read_base_r2_16s_ref_dict[ref_id][ref_pos] = cigar
                else:
                    ################################### analysis previous read refs ####################################

                    current_read_base___qualified_reads             = 0
                    current_read_base___consider_r1_unmapped_mate   = 0
                    current_read_base___consider_r2_unmapped_mate   = 0
                    current_read_base___both_mapped_to_16s          = 0
                    current_read_base___r1_16s_ref_dict             = dict()
                    current_read_base___r2_16s_ref_dict             = dict()
                    current_read_base___r1_16s_refs_no_ignored      = dict()
                    current_read_base___r2_16s_refs_no_ignored      = dict()
                    current_read_base___shared_16s_refs_no_ignored  = dict()

                    ########## get lowest mismatch for r1/r2 16s refs ##########

                    # get r1_ref_cigar_set
                    r1_ref_cigar_set = set()
                    for each_pos_dict in current_read_base_r1_16s_ref_dict.values():
                        each_pos_dict_values = {each_pos_dict[i] for i in each_pos_dict}
                        r1_ref_cigar_set.update(each_pos_dict_values)

                    # get r2_ref_cigar_set
                    r2_ref_cigar_set = set()
                    for each_pos_dict in current_read_base_r2_16s_ref_dict.values():
                        each_pos_dict_values = {each_pos_dict[i] for i in each_pos_dict}
                        r2_ref_cigar_set.update(each_pos_dict_values)

                    r1_ref_min_mismatch = get_min_mismatch_from_cigar_list(r1_ref_cigar_set, min_M_len_16s)
                    r2_ref_min_mismatch = get_min_mismatch_from_cigar_list(r2_ref_cigar_set, min_M_len_16s)

                    refs_to_ignore = set()

                    ########## filter r1 16s refs ##########

                    r1_16s_refs_passed_qc = {}
                    r1_16s_refs_passed_qc_with_pos = {}
                    for r1_16s_ref in current_read_base_r1_16s_ref_dict:
                        r1_matched_pos_dict = current_read_base_r1_16s_ref_dict[r1_16s_ref]

                        # one read need to mapped to one 16S only for one time
                        if len(r1_matched_pos_dict) > 1:
                            refs_to_ignore.add(r1_16s_ref)
                        else:
                            r1_16s_ref_pos = list(r1_matched_pos_dict.keys())[0]
                            r1_16s_ref_cigar = r1_matched_pos_dict[r1_16s_ref_pos]
                            r1_16s_ref_cigar_splitted = cigar_splitter(r1_16s_ref_cigar)

                            # check both end clip
                            both_end_clp = check_both_ends_clipping(r1_16s_ref_cigar_splitted)
                            if both_end_clp is True:
                                refs_to_ignore.add(r1_16s_ref)
                            else:
                                # check mismatch
                                r1_aligned_len, r1_aligned_pct, r1_clipping_len, r1_clipping_pct, r1_mismatch_pct = get_cigar_stats(
                                    r1_16s_ref_cigar_splitted)
                                if r1_ref_min_mismatch == 'NA':
                                    refs_to_ignore.add(r1_16s_ref)
                                elif (r1_mismatch_pct > r1_ref_min_mismatch) or (r1_mismatch_pct > mismatch_cutoff):
                                    refs_to_ignore.add(r1_16s_ref)
                                else:
                                    # check aligned length
                                    if r1_aligned_len < min_M_len_16s:
                                        refs_to_ignore.add(r1_16s_ref)
                                    else:
                                        # check if clp in the middle
                                        clip_in_middle = False
                                        if ('S' in r1_16s_ref_cigar) or ('s' in r1_16s_ref_cigar):
                                            clip_in_middle = True
                                            if (r1_16s_ref_cigar_splitted[0][-1] in ['S', 's']) and (
                                                    r1_16s_ref_pos == 1):
                                                clip_in_middle = False
                                            if (r1_16s_ref_cigar_splitted[-1][-1] in ['S', 's']):
                                                if (r1_16s_ref_pos + r1_aligned_len - 1) == marker_len_dict[r1_16s_ref]:
                                                    clip_in_middle = False

                                        # exclude the ref if clp in the middle is True
                                        if clip_in_middle is True:
                                            refs_to_ignore.add(r1_16s_ref)
                                        else:
                                            r1_16s_refs_passed_qc[r1_16s_ref] = [r1_16s_ref_cigar]
                                            r1_16s_refs_passed_qc_with_pos[r1_16s_ref] = {r1_16s_ref_pos: r1_16s_ref_cigar}

                    ########## filter r2 16s refs ##########

                    r2_16s_refs_passed_qc = {}
                    r2_16s_refs_passed_qc_with_pos = {}
                    for r2_16s_ref in current_read_base_r2_16s_ref_dict:
                        r2_matched_pos_dict = current_read_base_r2_16s_ref_dict[r2_16s_ref]

                        # one read need to mapped to one 16S only once
                        if len(r2_matched_pos_dict) > 1:
                            refs_to_ignore.add(r2_16s_ref)
                        else:
                            r2_16s_ref_pos = list(r2_matched_pos_dict.keys())[0]
                            r2_16s_ref_cigar = r2_matched_pos_dict[r2_16s_ref_pos]
                            r2_16s_ref_cigar_splitted = cigar_splitter(r2_16s_ref_cigar)

                            # check both end clip
                            both_end_clp = check_both_ends_clipping(r2_16s_ref_cigar_splitted)
                            if both_end_clp is True:
                                refs_to_ignore.add(r2_16s_ref)
                            else:
                                # check mismatch
                                r2_aligned_len, r2_aligned_pct, r2_clipping_len, r2_clipping_pct, r2_mismatch_pct = get_cigar_stats(
                                    r2_16s_ref_cigar_splitted)
                                if r2_ref_min_mismatch == 'NA':
                                    refs_to_ignore.add(r2_16s_ref)
                                elif (r2_mismatch_pct > r2_ref_min_mismatch) or (r2_mismatch_pct > mismatch_cutoff):
                                    refs_to_ignore.add(r2_16s_ref)
                                else:
                                    # check aligned length
                                    if r2_aligned_len < min_M_len_16s:
                                        refs_to_ignore.add(r2_16s_ref)
                                    else:
                                        # check if clp in the middle
                                        clip_in_middle = False
                                        if ('S' in r2_16s_ref_cigar) or ('s' in r2_16s_ref_cigar):
                                            clip_in_middle = True
                                            if (r2_16s_ref_cigar_splitted[0][-1] in ['S', 's']) and (
                                                    r2_16s_ref_pos == 1):
                                                clip_in_middle = False
                                            if (r2_16s_ref_cigar_splitted[-1][-1] in ['S', 's']):
                                                if (r2_16s_ref_pos + r2_aligned_len - 1) == marker_len_dict[r2_16s_ref]:
                                                    clip_in_middle = False

                                        # exclude the ref if clp in the middle is True
                                        if clip_in_middle is True:
                                            refs_to_ignore.add(r2_16s_ref)
                                        else:
                                            r2_16s_refs_passed_qc[r2_16s_ref] = [r2_16s_ref_cigar]
                                            r2_16s_refs_passed_qc_with_pos[r2_16s_ref] = {r2_16s_ref_pos: r2_16s_ref_cigar}

                    #################################

                    r1_16s_refs_no_ignored          = {key: value for key, value in r1_16s_refs_passed_qc.items() if key not in refs_to_ignore}
                    r2_16s_refs_no_ignored          = {key: value for key, value in r2_16s_refs_passed_qc.items() if key not in refs_to_ignore}
                    r1_16s_refs_no_ignored_with_pos = {key: value for key, value in r1_16s_refs_passed_qc_with_pos.items() if key not in refs_to_ignore}
                    r2_16s_refs_no_ignored_with_pos = {key: value for key, value in r2_16s_refs_passed_qc_with_pos.items() if key not in refs_to_ignore}

                    r1_ctg_refs_rd2_no_ignored_str_list = {('%s__cigar__%s' % (key, value[0])) for key, value in r1_16s_refs_passed_qc.items() if key not in refs_to_ignore}
                    r2_ctg_refs_rd2_no_ignored_str_list = {('%s__cigar__%s' % (key, value[0])) for key, value in r2_16s_refs_passed_qc.items() if key not in refs_to_ignore}

                    r1_ctg_refs_rd2_no_ignored_str_list_with_pos = set()
                    for each_16s in r1_16s_refs_no_ignored_with_pos:
                        cigar_dict = r1_16s_refs_no_ignored_with_pos[each_16s]
                        for each_cigar_pos in cigar_dict:
                            cigar_str = cigar_dict[each_cigar_pos]
                            r1_ctg_refs_rd2_no_ignored_str_list_with_pos.add( '%s__pc__%s__pc__%s' % (each_16s, each_cigar_pos, cigar_str))




                    print(r1_16s_refs_no_ignored)
                    print(r1_16s_refs_no_ignored_with_pos)
                    print(r1_ctg_refs_rd2_no_ignored_str_list)
                    print(r1_ctg_refs_rd2_no_ignored_str_list_with_pos)
                    print()





                    # print('\n----------------------------------\n')
                    #
                    # print('r1_16s_refs_passed_qc: %s' % r1_16s_refs_passed_qc)
                    # print('r1_16s_refs_passed_qc_with_pos: %s' % r1_16s_refs_passed_qc_with_pos)
                    # print()
                    # print('r1_16s_refs_no_ignored: %s' % r1_16s_refs_no_ignored)
                    # print('r1_16s_refs_no_ignored_with_pos: %s' % r1_16s_refs_no_ignored_with_pos)
                    # print()
                    # print('r2_16s_refs_passed_qc: %s' % r2_16s_refs_passed_qc)
                    # print('r2_16s_refs_passed_qc_with_pos: %s' % r2_16s_refs_passed_qc_with_pos)
                    # print()
                    # print('r2_16s_refs_no_ignored: %s' % r2_16s_refs_no_ignored)
                    # print('r2_16s_refs_no_ignored_with_pos: %s' % r2_16s_refs_no_ignored_with_pos)


                    # no mate has no_ignored alignments
                    if (len(r1_16s_refs_no_ignored) == 0) and (len(r2_16s_refs_no_ignored) == 0):
                        pass

                    # only r1 has no_ignored alignments
                    elif (len(r1_16s_refs_no_ignored) > 0) and (len(r2_16s_refs_no_ignored) == 0):

                        current_read_base___qualified_reads = 1
                        current_read_base___consider_r1_unmapped_mate = 1
                        current_read_base___r1_16s_refs_no_ignored = r1_16s_refs_no_ignored
                        current_read_base___r1_16s_ref_dict = current_read_base_r1_16s_ref_dict

                    # only r2 has no_ignored alignments
                    elif (len(r1_16s_refs_no_ignored) == 0) and (len(r2_16s_refs_no_ignored) > 0):

                        current_read_base___qualified_reads = 1
                        current_read_base___consider_r2_unmapped_mate = 1
                        current_read_base___r2_16s_refs_no_ignored = r2_16s_refs_no_ignored
                        current_read_base___r2_16s_ref_dict = current_read_base_r2_16s_ref_dict

                    # both r1 and r2 have no_ignored alignments
                    else:
                        shared_16s_ref_dict          = {key: [r1_16s_refs_no_ignored[key][0], r2_16s_refs_no_ignored[key][0]] for key in set(r1_16s_refs_no_ignored).intersection(set(r2_16s_refs_no_ignored))}
                        shared_16s_ref_dict_with_pos = {key: [r1_16s_refs_no_ignored[key][0], r2_16s_refs_no_ignored[key][0]] for key in set(r1_16s_refs_no_ignored).intersection(set(r2_16s_refs_no_ignored))}

                        if len(shared_16s_ref_dict) > 0:
                            # print('\n-----\n')
                            # print(r1_16s_refs_no_ignored)
                            # print(r2_16s_refs_no_ignored)
                            # print(shared_16s_ref_dict)
                            # print(shared_16s_ref_dict_with_pos)
                            # print(r1_16s_refs_no_ignored_with_pos)
                            # print(r2_16s_refs_no_ignored_with_pos)
                            # print(shared_16s_ref_dict_with_pos)
                            current_read_base___qualified_reads = 1
                            current_read_base___both_mapped_to_16s = 1
                            current_read_base___shared_16s_refs_no_ignored = shared_16s_ref_dict
                            current_read_base___r1_16s_ref_dict = current_read_base_r1_16s_ref_dict
                            current_read_base___r2_16s_ref_dict = current_read_base_r2_16s_ref_dict

                    if current_read_base___qualified_reads == 1:

                        # print('\n----------------------------------\n')
                        # print('current_read_base___r1_16s_ref_dict: %s' % current_read_base___r1_16s_ref_dict)
                        # print('current_read_base___r2_16s_ref_dict: %s' % current_read_base___r2_16s_ref_dict)
                        # print('current_read_base___r1_16s_refs_no_ignored: %s' % current_read_base___r1_16s_refs_no_ignored)
                        # print('current_read_base___r2_16s_refs_no_ignored: %s' % current_read_base___r2_16s_refs_no_ignored)
                        # print('current_read_base___shared_16s_refs_no_ignored: %s' % current_read_base___shared_16s_refs_no_ignored)


                        current_read_base___r1_16s_ref_dict_str            = r12_16s_ref_dict_to_str(current_read_base___r1_16s_ref_dict)
                        current_read_base___r2_16s_ref_dict_str            = r12_16s_ref_dict_to_str(current_read_base___r2_16s_ref_dict)
                        current_read_base___r1_16s_refs_no_ignored_str     = no_ignored_dict_to_str(current_read_base___r1_16s_refs_no_ignored)
                        current_read_base___r2_16s_refs_no_ignored_str     = no_ignored_dict_to_str(current_read_base___r2_16s_refs_no_ignored)
                        current_read_base___shared_16s_refs_no_ignored_str = no_ignored_dict_to_str(current_read_base___shared_16s_refs_no_ignored)
                        MappingRecord_file_handle.write('%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n' % (current_read_base,
                                                                                                      current_read_base___qualified_reads,
                                                                                                      current_read_base___consider_r1_unmapped_mate,
                                                                                                      current_read_base___consider_r2_unmapped_mate,
                                                                                                      current_read_base___both_mapped_to_16s,
                                                                                                      current_read_base___r1_16s_ref_dict_str,
                                                                                                      current_read_base___r2_16s_ref_dict_str,
                                                                                                      current_read_base___r1_16s_refs_no_ignored_str,
                                                                                                      current_read_base___r2_16s_refs_no_ignored_str,
                                                                                                      current_read_base___shared_16s_refs_no_ignored_str))


                    ########################################### reset values ###########################################

                    current_read_base = read_id_base
                    current_read_base_r1_16s_ref_dict = dict()
                    current_read_base_r2_16s_ref_dict = dict()

                    if cigar != '*':
                        if read_strand == '1':
                            current_read_base_r1_16s_ref_dict[ref_id] = {ref_pos: cigar}
                        if read_strand == '2':
                            current_read_base_r2_16s_ref_dict[ref_id] = {ref_pos: cigar}

    MappingRecord_file_handle.close()


pwd_splitted_sam_file       = '/Users/songweizhi/Desktop/666/subset.sam'
input_16s_qc                = '/Users/songweizhi/Desktop/666/CAMI_Oral_138_16S_0.999.polished_min1200.QC.fa'
pwd_splitted_sam_mp_file    = '/Users/songweizhi/Desktop/666/subset_mp.txt'
min_M_len_16s               = 45
mismatch_cutoff             = 2

# head -100000 Oral_0715_60_60_min1200_input_reads_to_16S_sorted.sam > subset.sam

marker_len_dict = {}
for each_marker_record in SeqIO.parse(input_16s_qc, 'fasta'):
    marker_len_dict[each_marker_record.id] = len(each_marker_record.seq)

arg_list = [pwd_splitted_sam_file, pwd_splitted_sam_mp_file,
            min_M_len_16s, mismatch_cutoff, marker_len_dict]

parse_sam16s_worker(arg_list)





def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def continuous_mismatch_cigar(cigar_splitted):

    print(cigar_splitted)

    if cigar_splitted[0][-1] in ['S', 's']:
        cigar_splitted = cigar_splitted[1:]

    if cigar_splitted[-1][-1] in ['S', 's']:
        cigar_splitted = cigar_splitted[:-1]

    print(cigar_splitted)


    mismatch_len_list = []
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]
        if (each_part_cate in ['=', 'S']) and (each_part_len > 10):
            if mismatch_len > 0:
                mismatch_len_list.append(mismatch_len)
            mismatch_len = 0
        else:
            if each_part_cate not in ['=', 'S']:
                mismatch_len += each_part_len

    print(mismatch_len_list)

    continuous_mismatch = False
    if max(mismatch_len_list) > 1:
        continuous_mismatch = True

    print(max(mismatch_len_list))
    return continuous_mismatch


'''
22=2X65=
22=1X1D65=
22=1X1I65=
22=2D65=
'''

cigar = '6S8=1D2=1X65S'

print(continuous_mismatch_cigar(cigar_splitter(cigar)))





blast_best_hits = '/Users/songweizhi/Desktop/BH_ER_050417_vs_GTDB_r95_BestHit.tab'
ref_taxon_file  = '/Users/songweizhi/Desktop/GTDB_ssu_all_r95.txt'
query_taxon     = '/Users/songweizhi/Desktop/query_taxon.txt'

ref_taxon_dict = {}
for each_ref_seq in open(ref_taxon_file):
    each_ref_seq_split = each_ref_seq.strip().split(' ')
    ref_seq_id = each_ref_seq_split[0]
    ref_seq_taxon = ' '.join(each_ref_seq_split[1:])
    ref_taxon_dict[ref_seq_id] = ref_seq_taxon


query_to_hit_dict = {}
for each_hit in open(blast_best_hits):
    each_hit_split = each_hit.strip().split('\t')
    query_to_hit_dict[each_hit_split[0]] = each_hit_split[1]


query_taxon_handle = open(query_taxon, 'w')
for each_query in query_to_hit_dict:
    query_taxon = ref_taxon_dict[query_to_hit_dict[each_query]]
    query_taxon_handle.write('%s\t%s\n' % (each_query, query_taxon))
query_taxon_handle.close()




